import numpy as np
import pandas as pd
import matplotlib.pyplot as plt
from rotate import Varimax, Promax
from PlotComp import PlotPattern
from ArrangePat import ArrangePtn

"""
                   Direct Least Squares Method
                      for  Factor Analysis

                     Yasuharu Okamoto, 2023.11

"""
#
#       Excelデータファイルの読み込み
#
s = input("Input Excel data file (*.xlsx) = ")
xlsx_fl = pd.ExcelFile(s)
data_xlsx = pd.read_excel(xlsx_fl,  header = None)
data = data_xlsx.values

f_out_nm = input('Output Text file(*.txt) = ')
f_out = open(f_out_nm, 'w')

print('data =\n', data)
f_out.write(f'\nInput data = {f_out_nm}\n')
f_out.write(f'\ndata =\n{data}\n')

var_names = data[0]
data_id = data.T[0][1:]
X = data[1:][:,1:]
X = np.array(X, dtype='float')
n, n_vars = np.shape(X)
print('var_names =\n', var_names)
print('data_id =\n', data_id)
print('n =', n)
print('n_vars =', n_vars)
print('X =\n', X)
var_means = np.mean(X, axis=0)
print('var_means =\n', var_means)
var_sds = np.std(X, axis=0)
print('var_sds =\n', var_sds)
X = (X - var_means) / var_sds
print('X-means =\n', np.sum(X,axis=0) / n)
print('X-sds =\n', np.sum(X**2, axis=0) / n)

U, S, Vh = np.linalg.svd(X, full_matrices = False)   #   特異値分解

print('U =\n', U)
print('S =\n', S)
print('Vh =\n', Vh)
ck_idx = S > 1.0e-7
print('ck_idx =\n', ck_idx)
U = U[:, ck_idx]
print('U =\n', U)
S = S[ck_idx]
print('S =\n', S)
Vh = Vh[ck_idx]
print('Vh =\n', Vh)

print('Singular values =\n', S)

plt.plot(np.arange(len(S)), S)
plt.xticks(np.arange(len(S)), np.arange(1, len(S) + 1))
plt.ylabel('Singular value', fontsize = 12)
plt.show()

p = int(input('Number of factors = '))
if p > len(X[0]):
    p = len(X[0])

A = Vh.T[:,:p] @ np.diag(S[:p]) / (n**0.5)
print('A =\n', A)

PlotPtn = PlotPattern(A, var_names[1:], n_vars, p, "Initial Pattern Matrix" )
while True:
    print("\nSet the factor number (from 1 to %d) to plot the pattern."%p)
    print("If you do not want to plot the pattern, set the number less than 1.")
    xdim = int(input("xdim = "))
    if xdim < 1:
        break
    ydim = int(input("ydim = "))
    PlotPtn.Plot(xdim, ydim)

commu = np.sum(A**2, axis=1)
             
f_out.write("\n\nPattern matrix and communalities for the initial factors\n")
f_out.write("\n%10s"%" ")
for i in range(0, p):
    f_out.write("%10s"%("fctr.%d"%(i + 1)))
f_out.write("%15s\n"%"communality")
for i in range(0, n_vars):
    f_out.write("%10s"%var_names[i+1])
    for j in range(0, p):
        f_out.write("%10.5f"%A[i, j])
    f_out.write("%15.5f\n"%commu[i])


rot = Varimax(A, n_vars, p)      #    Varimax回転
PatV = rot.rotateV()
            
    
PlotPtn = PlotPattern( PatV, var_names[1:], n_vars, p, "Pattern Matrix(Varimax)" )
while True:
    print("\nSet the factor number (from 1 to %d) to plot the pattern for the Varimax criterion."%p)
    print("If you do not want to plot the pattern, set the number less than 1.")
    xdim = int(input("xdim = "))
    if xdim < 1:
        break
    ydim = int(input("ydim = "))
    PlotPtn.Plot(xdim, ydim)


            
f_out.write("\n\nVarimax rotation was applied...\n\n")
f_out.write("\nPattern for the Varimax rotated factors\n\n")
f_out.write("%10s"%" ")
for j in range(0, p):
    f_out.write("%10s"%("fctr.%d"%(j + 1)))
f_out.write("\n")
for i in range(0, n_vars):
    f_out.write("%10s"%var_names[i+1])
    for j in range(0, p):
        f_out.write("%10.5f"%PatV[i, j])
    f_out.write("\n")

ArgPtn = ArrangePtn(var_names[1:], PatV, n_vars, p)
PtnArgd = ArgPtn.arrange()
f_out.write("\n\nSorted Pattern for Varimax rotation\n\n")
f_out.write("%10s"%" ")
for j in range(0, p):
    f_out.write("%10s"%("fctr.%d"%(j + 1)))
f_out.write("\n")
for i in range(0, n_vars):
    f_out.write("%10s"%PtnArgd[i][0])
    for j in range(0, p):
        f_out.write("%10.5f"%PtnArgd[i][1][j])
    f_out.write("\n")
    



obrot = Promax(PatV, n_vars, p, 4.0)     #   Promax回転
Patob, Phi = obrot.rotateP();

f_out.write("\n\nPromax rotation was applied...\n\n")
f_out.write("\nPattern for the Promax rotated factors\n\n")
f_out.write("%10s"%" ")
for j in range(0, p):
    f_out.write("%10s"%("fctr.%d"%(j + 1)))
f_out.write("\n")

for i in range(0, n_vars):
    f_out.write("%10s"%var_names[i+1])
    for j in range(0, p):
        f_out.write("%10.5f"%Patob[i, j])
    f_out.write("\n")

f_out.write("\n\nCorrelation matrix of the factors\n\n")
f_out.write("%10s"%" ")
for j in range(0, p):
    f_out.write("%10s"%("fctr.%d"%(j + 1)))
f_out.write("\n")
for i in range(0, p):
    f_out.write("%10s"%("fctr.%d"%(i + 1)))
    for j in range(0, p):
        f_out.write("%10.5f"%Phi[i, j])
    f_out.write("\n")

ArgPtn = ArrangePtn(var_names[1:], Patob, n_vars, p)
PtnArgd = ArgPtn.arrange()
f_out.write("\n\nSorted Pattern for Promax rotation\n\n")
f_out.write('{0:10s}'.format(' '))
for i in range(p):
    str = 'fctr.{}'.format(i + 1)
    f_out.write('{0:>10s}'.format(str))
f_out.write('\n')
for i in range(0, n_vars):
    f_out.write("%10s"%PtnArgd[i][0])
    for j in range(0, p):
        f_out.write("%10.5f"%PtnArgd[i][1][j])
    f_out.write("\n")

S = Patob * Phi
f_out.write('\n\nStructure Matrix\n\n')
f_out.write('{0:10s}'.format(' '))
for i in range(p):
    str = 'fctr.{}'.format(i + 1)
    f_out.write('{0:>10s}'.format(str))
f_out.write('\n')
for i in range(n_vars):
    f_out.write('{0:>10s}'.format(var_names[i+1]))
    for j in range(p):
        f_out.write('{0:10.5f}'.format(S[i, j]))
    f_out.write('\n')


f_out.close()
print(f'\n{f_out_nm} was saved.\n')


